/***************************************************************************
                          dcchat.cpp  -  description
                             -------------------
    begin                : Don Mr 28 2002
    copyright            : (C) 2002 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qdatetime.h>
#include <qlineedit.h>
#include <qpushbutton.h>
#include <qclipboard.h>
#include <qapplication.h>
#include <qmessagebox.h>
#include <qimage.h>
#include <qregexp.h>
#include <qpopupmenu.h>
#include <qcursor.h>
#include <qcombobox.h>
#include <qlayout.h>
#include <qprocess.h>
#include <qglobal.h>
#include <qstatusbar.h>
#include <qtooltip.h>
#include <qfile.h>
#include <qfiledialog.h>
#include <qsplitter.h>
#include <qtextcodec.h>
#include <qinputdialog.h>

#include <dcgui.h>
#include <dcevent.h>
#include <dcclient.h>
#include <dcmenuhandler.h>
#include <dcconfig.h>
#include <dctranslator.h>
#include <cdialogpicturemap.h>
#include <dcfiletool.h>
#include <dcserverlist.h>
#include <dcuserslist.h>

#include "dcchat.h"

#include <dclib/cfilemanager.h>
#include <dclib/dcos.h>
#include <dclib/cutils.h>
#include <dclib/core/cbase64.h>
#include <dclib/core/cdir.h>
#include <dclib/core/cfile.h>

//#include <res/encrypted.xpm>
//#include <res/decrypted.xpm>
#include <res/ssl_no.xpm>
#include <res/ssl_yes.xpm>

/** the link list*/
const char * p_LinkList[] = {
	"HTTP://",
	"HTTPS://",
	"NEWS://",
	"FILE://",
	"WWW.",
	"FTP://",
	"FTP.",
	"HKP://",
	"LDAP://",
	"DCFILE://",
	"DCHUB://",
	NULL
};

/** */
DCChat::DCChat( QWidget * parent, const char *name, int wflags, DCClient * client, bool bprivate ) : DCDialogChat(parent, name, wflags)
{
	ASSERT(client);

	m_bPrivateChat = bprivate;
	m_pClient      = client;
	bSendAway      = TRUE;
	m_eSecureState = esecsNONE;
	m_pCurrentHistory = 0;

	m_sTimeStamp  = QDate::currentDate().toString("yyyyMMdd").ascii();
	m_sTimeStamp += "-";
	m_sTimeStamp += QTime::currentTime().toString("hhmmss").ascii();

	InitDocument();
}

/** */
DCChat::~DCChat()
{
}

/** */
void DCChat::InitDocument()
{
	int i;

	m_pStatusBar = new QStatusBar(this);
	DCDialogChatLayout->addWidget( m_pStatusBar, 1, 0 );
	m_pStatusCrypt = new QLabel(m_pStatusBar);
	m_pStatusBar->addWidget(m_pStatusCrypt);

	SetCrypt(esecsNONE);

	if ( !m_bPrivateChat )
	{
		m_pStatusBar->hide();
	}
	else
	{
		m_pStatusBar->show();
	}

/*	if ( Config->ShowChatStatusBar() == FALSE )
	{
		m_pStatusBar->hide();
	}
*/

	if ( Config->GetShowChatSendButton() )
	{
		connect( PushButton_SEND, SIGNAL(clicked()), this, SLOT(slotSendMessage()) );
		slotTextChangedChatInput();
	}
	else
	{
		// remove send button and recreate layout
		delete PushButton_SEND;
		PushButton_SEND = NULL;
		delete Frame_SEND->layout();
		QGridLayout * gl = new QGridLayout( Frame_SEND, 1, 1, 1, 1, "");
		gl->addWidget( TextEdit_CHATINPUT, 0, 0 );
	}

	TextEdit_CHATOUTPUT = new DCQTextEdit( Frame_OUTPUT, "TextEdit_CHATOUTPUT" );
	TextEdit_CHATOUTPUT->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)5, (QSizePolicy::SizeType)5, 0, 0, TextEdit_CHATOUTPUT->sizePolicy().hasHeightForWidth() ) );
	TextEdit_CHATOUTPUT->setMinimumSize( QSize( 0, 60 ) );
	TextEdit_CHATOUTPUT->setTextFormat( QTextEdit::RichText );
	TextEdit_CHATOUTPUT->setWordWrap( QTextEdit::WidgetWidth );
	TextEdit_CHATOUTPUT->setWrapPolicy( QTextEdit::AtWhiteSpace );
    	TextEdit_CHATOUTPUT->setReadOnly( TRUE );
	Frame_OUTPUTLayout->addWidget(TextEdit_CHATOUTPUT,0,0);

	connect( TextEdit_CHATOUTPUT, SIGNAL(rightButtonClicked( const QPoint& )), this, SLOT(slotRightButtonClickedChatOutput( const QPoint& )) );
#if QT_VERSION >= 0x030100
	connect( TextEdit_CHATOUTPUT, SIGNAL(clicked( int, int )), this, SLOT(slotClickedChatOutput( int, int )) );
#endif
	if ( PushButton_SEND )
	{
		connect( TextEdit_CHATINPUT, SIGNAL(textChanged()), this, SLOT(slotTextChangedChatInput()) );
	}

	installEventFilter(this);

	TextEdit_CHATOUTPUT->installEventFilter(this);
	TextEdit_CHATINPUT->installEventFilter(this);
	TextEdit_CHATINPUT->setFocus();

	// resize splitter
	QValueList<int> list = Splitter3->sizes();
	QValueList<int>::Iterator it = list.begin();
	i = *it;
	++it;
	i += *it;
	i -= TextEdit_CHATINPUT->height();
	it = list.begin();
	*it = i;
	++it;
	*it = TextEdit_CHATINPUT->height();
	Splitter3->setSizes(list);
}

/** event filter */
bool DCChat::eventFilter( QObject * object, QEvent * event )
{
	if ((event->type() == QEvent::KeyPress)&&((QTextEdit*)object==TextEdit_CHATINPUT))
	{
		QKeyEvent * e = (QKeyEvent*)event;

		bool send = FALSE;
		if ( e->state() == AltButton )
		{
			if ( (e->key() == Key_S) && (Config->GetSendChat() == "Alt + S") )
			{
				send = TRUE;
			}
			else if( ((e->key() == Key_Enter) || (e->key() == Key_Return )) &&
				   (Config->GetSendChat() == "Alt + Enter"))
			{
				send = TRUE;
			}
		}
		else if( (Config->GetSendChat() == "Ctrl + Enter") &&
			 (e->state() == ControlButton) && ( e->key() == Key_Enter || e->key() == Key_Return) )
		{
			send = TRUE;
		}
		else if( (Config->GetSendChat() == "Enter" && (e->state() != ControlButton) &&
			 ( e->key() == Key_Enter || e->key() == Key_Return)) )
		{
			send = TRUE;
		}

		if( send == TRUE )
		{
			SendMessage();
			e->accept();
			return true;
		}
		else if ( e->state() == ControlButton )
		{
			CString * ps = 0;

			// check for history
			if ( (e->key() == Key_Down) && (m_pCurrentHistory != 0) )
			{
				ps = m_History.Next(m_pCurrentHistory);
			}
			else if ( e->key() == Key_Up )
			{
				ps = m_History.Forw(m_pCurrentHistory);
			}

			if ( ps != 0 )
			{
				m_pCurrentHistory = ps;
				TextEdit_CHATINPUT->setText(m_pCurrentHistory->Data());
			}
			else if (e->key() == Key_Down)
			{
				m_pCurrentHistory = ps;
				TextEdit_CHATINPUT->clear();
			}
		}
	}
	else if ( event->type() == EVENT_TRANSLATION )
	{
		DC_TranslationEvent *e = (DC_TranslationEvent*)event;

		if ( e->m_bTranslate == TRUE )
		{
			AddStatus( (tr("Translation: ") + "'" + e->m_sOriginal + "' -> '" + e->m_sTranslation + "'").ascii() );
		}
		else
		{
			AddStatus( (tr("Translation failed: ") + "'" + e->m_sOriginal + "'").ascii() );
		}
	}

	return QWidget::eventFilter( object, event );    // standard event processing
}

/** */
void DCChat::SetCrypt( eSecureState e )
{
	if ( e == esecsENCRYPTED )
	{
		QToolTip::remove(m_pStatusCrypt);
		QToolTip::add(m_pStatusCrypt, tr("Line is encrypted."));
		m_pStatusCrypt->setPixmap( QPixmap( ssl_yes_xpm ));
	}
	else
	{
		QToolTip::remove(m_pStatusCrypt);
		QToolTip::add(m_pStatusCrypt, tr("Line is not encrypted."));
		m_pStatusCrypt->setPixmap( QPixmap( ssl_no_xpm ));
	}

	m_eSecureState = e;
}

/** */
void DCChat::slotTextChangedChatInput()
{
	if ( PushButton_SEND )
	{
		if ( TextEdit_CHATINPUT->text() != "" )
		{
			PushButton_SEND->setEnabled(TRUE);
		}
		else
		{
			PushButton_SEND->setEnabled(FALSE);
		}
	}
}

/** */
void DCChat::slotClickedChatOutput( int, int )
{
	int i;
	CString pressedLink;
	QPoint p;
	int x,y;
	DCFileTool filetool;

	p = TextEdit_CHATOUTPUT->mapFromGlobal(QCursor::pos());
	TextEdit_CHATOUTPUT->viewportToContents(p.x(),p.y(),x,y);
	p.setX(x);
	p.setY(y);
	pressedLink = TextEdit_CHATOUTPUT->anchorAt(p).ascii();

//	printf("%s\n",pressedLink.ascii());

	for(i=0;p_LinkList[i]!=0;i++)
	{
		if ( pressedLink.Left(CString(p_LinkList[i]).Length()).ToUpper() == p_LinkList[i] )
		{
			if ( p_LinkList[i] == CString("DCFILE://") )
			{
				// DC-File-Link
				QPopupMenu *m;
				int id;
				CString hubhost,hubname,nick,file;
				ulonglong size;

				if ( CUtils::ConvertDCLink( pressedLink, hubhost, hubname, nick, size, file ) == FALSE )
				{
					return;
				}

				m = new QPopupMenu(this);

				DCMenuHandler::InsertMenu( m, emiDOWNLOAD );
				DCMenuHandler::InsertMenu( m, emiDOWNLOAD_TO );
				DCMenuHandler::InsertMenu( m, emiDOWNLOAD_AS );
				DCMenuHandler::InsertMenu( m, emiDOWNLOAD_IN );

				id = m->exec(QCursor::pos());

				delete m;

				if ( (id == emiDOWNLOAD) || (id == emiDOWNLOAD_AS) || (id == emiDOWNLOAD_TO) )
				{
					QString localrootpath = "";
					QString localname = "";

					// select downloadfolder for all selected files
					if ( id == emiDOWNLOAD_TO )
					{
						localrootpath = QFileDialog::getExistingDirectory( "", this, "bdf", tr("Select download folder"), TRUE );

						if ( localrootpath == "" )
							return;
					}

					QFileInfo fi(file.Data());
					localname = fi.fileName();

					if ( id == emiDOWNLOAD_AS )
					{
						localrootpath = QFileDialog::getSaveFileName( "", "", this, "bdf", tr("Select file for")+" "+localname );

						if ( localrootpath == "" )
							return;

						QFileInfo fi(localrootpath);
						localrootpath = fi.dirPath();
						localname     = fi.fileName();

						if ( (localrootpath=="") || (localname=="") )
							return;
					}

					// add transfer to the waitlist
					filetool.CheckFile( nick, hubname, hubhost,
							file, localname.ascii(), "", localrootpath.ascii(), eltFILE,
							size );
				}
				else if ( id == emiDOWNLOAD_IN )
				{
					QString localrootpath;
					QString localname;
					QString localpath;

					if ( filetool.SelectFileSource( size, localname, localrootpath, localpath ) == FALSE )
					{
						return;
					}

					// add transfer to the waitlist
					filetool.CheckFile( nick, hubname, hubhost,
							  file, localname.ascii(), localpath.ascii(), localrootpath.ascii(), eltFILE,
					  		size, TRUE);
				}
			}
			else if ( p_LinkList[i] == CString("DCHUB://") )
			{
				pressedLink = pressedLink.Right( pressedLink.Length()-8 );
				// remove all '/' chars
				pressedLink = pressedLink.Replace('/',"");

				pServerList->Connect( pressedLink, pressedLink );
			}
			else if ( Config->GetBrowser() != "" )
			{
				QProcess * pr = new QProcess();
				QByteArray buf;
				pr->addArgument(Config->GetBrowser().Data());
				pr->addArgument(pressedLink.Data());
				pr->launch(buf);
				delete pr;
			}
		}
	}
}

/** */
void DCChat::slotRightButtonClickedChatOutput( const QPoint& )
{
	int id;

	QPopupMenu *m,*menc;
	QString s;

	m = new QPopupMenu(this);

	DCMenuHandler::InsertMenu( m, emiINSERTSMILEY, Config->GetEnableEmoticons() );
	DCMenuHandler::InsertMenu( m, emiSEPARATOR );
	DCMenuHandler::InsertMenu( m, emiCOPY, TextEdit_CHATOUTPUT->hasSelectedText() );
	DCMenuHandler::InsertMenu( m, emiCLEAR, TRUE );
	DCMenuHandler::InsertMenu( m, emiSELECT_ALL, TRUE );
	DCMenuHandler::InsertMenu( m, emiSEPARATOR );
	DCMenuHandler::InsertMenu( m, emiTRANSLATE, TextEdit_CHATOUTPUT->hasSelectedText() );
	DCMenuHandler::InsertMenu( m, emiTRANSLATOR );

	// encoding menu
	menc = DCMenuHandler::InsertMenu( m, emisTEXT_ENCODING );
	DCMenuHandler::InsertMenu( menc, emiTEXT_INPUT_ENCODING );
	DCMenuHandler::InsertMenu( menc, emiTEXT_OUTPUT_ENCODING );

	DCMenuHandler::InsertMenu( m, emiSEPARATOR );
	DCMenuHandler::InsertMenu( m, emiREFRESH );
	DCMenuHandler::InsertMenu( m, emiSEPARATOR );

	if ( m_bPrivateChat )
	{
		// insert request/close secure chat
		if ( m_eSecureState == esecsNONE )
			DCMenuHandler::InsertMenu( m, emiREQUEST_SECURE_CHAT );
		else
			DCMenuHandler::InsertMenu( m, emiCLOSE_SECURE_CHAT );
	}

	DCMenuHandler::InsertMenu( m, emiSAVE );

	if ( m_bPrivateChat )
	{
		DCMenuHandler::InsertMenu( m, emiHIDE, m_bPrivateChat && !Config->GetShowChatInTab() );
		DCMenuHandler::InsertMenu( m, emiCLOSE, m_bPrivateChat );
	}

	id = m->exec(QCursor::pos());

	delete m;

	if ( id == emiINSERTSMILEY )
	{
		int x,y;
		CList<DC_EmoticonObject> * elist;

		elist = Config->EmoticonList();

		if ( elist )
		{
			CDialogPictureMap * dialog = new CDialogPictureMap(this);
			QPixmap p(Config->GetEmoticonImage());
			dialog->SetPixmap(p);

			if ( dialog->exec() == QDialog::Accepted )
			{
				dialog->GetXY(x,y);

				DC_EmoticonObject * EmoticonObject = 0;

				while ( (EmoticonObject=elist->Next(EmoticonObject)) != 0 )
				{
					if ( ((EmoticonObject->left<x) && (EmoticonObject->right>x)) &&
					     ((EmoticonObject->top<y)  && (EmoticonObject->bottom>y)) )
					{
						TextEdit_CHATINPUT->insert(EmoticonObject->m_Text.Data());
						break;
					}
				}
			}

			delete dialog;
		}
	}
	else if ( id == emiTRANSLATE )
	{
		gDCTranslator->Translate( this, m_sLanguage.Data(), TextEdit_CHATOUTPUT->selectedText() );
	}
	else if ( id == emiTRANSLATOR )
	{
		m_sLanguage = gDCTranslator->SelectLanguage(m_sLanguage);
	}
	else if ( (id == emiTEXT_INPUT_ENCODING) ||
		  (id == emiTEXT_OUTPUT_ENCODING) )
	{
		QTextCodec *codec;
		int i;
		bool ok = FALSE;
		QStringList lst;
		QString res;

		for (i=0;(codec = QTextCodec::codecForIndex(i));i++)
		{
			lst << codec->name();
		}

		if ( id == emiTEXT_INPUT_ENCODING )
			i = lst.findIndex(Config->GetChatTextCodec(ectcINPUT).Data());
		else
			i = lst.findIndex(Config->GetChatTextCodec(ectcOUTPUT).Data());

		if ( i == -1 )
		{
			i = 0;
		}

		res = QInputDialog::getItem( tr("Text Encoding"), tr("Choose a codec"), lst, i, FALSE, &ok, 0 );

		if ( ok )
		{
			if ( QTextCodec::codecForName(res) )
			{
				if ( id == emiTEXT_INPUT_ENCODING )
					Config->SetChatTextCodec( ectcINPUT, res.ascii() );
				else
					Config->SetChatTextCodec( ectcOUTPUT, res.ascii() );
			}
		}
	}
	else if ( id == emiCOPY )
	{
		TextEdit_CHATOUTPUT->copy();
	}
	else if ( id == emiCLEAR )
	{
		TextEdit_CHATOUTPUT->clear();
		AddStatus( tr("Chat Cleared.").ascii() );
	}
	else if ( id == emiSELECT_ALL )
	{
		TextEdit_CHATOUTPUT->selectAll();
	}
	else if ( id == emiREQUEST_SECURE_CHAT )
	{
		m_pClient->SendPrivateMessage( m_pClient->GetNick(), m_sNick, "<request secchannel>" );
	}
	else if ( id == emiCLOSE_SECURE_CHAT )
	{
		m_pClient->SendPrivateMessage( m_pClient->GetNick(), m_sNick, "<close secchannel>" );
	}
	else if ( id == emiSAVE )
	{
		QString s = QFileDialog::getSaveFileName(
			"",
			"HTML (*.html *.htm)",
			this,
			"save file dialog",
			tr("Choose a filename to save under") );

		if ( s != "" )
		{
			QFile f(s);

			if ( f.open(IO_WriteOnly) == TRUE )
			{
				f.writeBlock(TextEdit_CHATOUTPUT->text(),TextEdit_CHATOUTPUT->text().length());
				f.close();
			}
		}
	}
	else if ( id == emiREFRESH )
	{
		if ( CFileManager::Instance()->CreateShareList() )
			s = tr("Refresh share in progress.");
		else
			s = tr("Refresh share allready in progress.");
		AddStatus(s.ascii());
	}
	else if ( id == emiCLOSE )
	{
		close();
	}
	else if ( id == emiHIDE )
	{
		setEnabled(FALSE);
		hide();
	}
}

/** */
bool DCChat::close( bool alsoDelete )
{
	// re-set the wflags ...
	setWFlags(WDestructiveClose);
	return QWidget::close(alsoDelete);
}

/** */
void DCChat::slotSendMessage()
{
	SendMessage();
}

/** */
void DCChat::SetNick( CString nick, CString hubname )
{
	m_sNick = nick;

	if ( nick != "" )
	{
		setCaption(tr("Private Chat:")+" "+nick.Data()+" ["+hubname.Data()+"]");
	}
}

/** */
CString DCChat::GetTimeStamp()
{
	CString s = "";

	if ( ((Config->GetTimeStamp(etsPRIVATECHAT) == TRUE) && (m_bPrivateChat)) ||
	     ((Config->GetTimeStamp(etsHUBCHAT) == TRUE) && (!m_bPrivateChat)) )
	{
		s = QTime::currentTime().toString("[hh:mm:ss]").ascii();
	}

	if ( m_bPrivateChat )
		s = "<font color=\"" + Config->GetChatColor(eccCHATTIMESTAMP) + "\">" + s + "</font> ";
	else
		s = "<font color=\"" + Config->GetChatColor(eccPUBLICCHATTIMESTAMP) + "\">" + s + "</font> ";

	return s;
}

/** */
void DCChat::AddOutput( CString message )
{
	bool bscroll;
	CString s;
	QTextCodec *codec;

	if ( TextEdit_CHATOUTPUT->verticalScrollBar()->maxValue() == TextEdit_CHATOUTPUT->verticalScrollBar()->value() )
	{
		bscroll = TRUE;
	}
	else
	{
		bscroll = FALSE;
	}

	// convert newline
	message = message.Replace( "\xa\\r", "\xa" );
	message = message.Replace( "\xa", "<br />" );

	if ( (Config->GetChatMaxParagraph()!= 0) && (TextEdit_CHATOUTPUT->paragraphs() > (Config->GetChatMaxParagraph()+5)) )
	{
		TextEdit_CHATOUTPUT->removeParagraph(0);
		TextEdit_CHATOUTPUT->removeParagraph(0);
		TextEdit_CHATOUTPUT->removeParagraph(0);
		TextEdit_CHATOUTPUT->removeParagraph(0);
		TextEdit_CHATOUTPUT->removeParagraph(0);
	}

	// use text encoder
	s = Config->GetChatTextCodec( ectcOUTPUT );

	if ( (s != "") && ( ( codec = QTextCodec::codecForName(s.Data())) != 0) )
	{
		TextEdit_CHATOUTPUT->append( codec->toUnicode(message.Data()) );
	}
	else
	{
		TextEdit_CHATOUTPUT->append( message.Data() );
	}

	if ( bscroll )
	{
		TextEdit_CHATOUTPUT->scrollToBottom();
		TextEdit_CHATOUTPUT->moveCursor( QTextEdit::MoveEnd, FALSE );
	}

#if QT_VERSION >= 304
	message += "<br />";
#endif

	// save to logfile
	if ( Config->GetLogChatOption(elcoENABLELOGGING) )
	{
		if ( !(Config->GetLogChatOption(elcoDISABLEPUBLICCHAT) && !m_bPrivateChat) )
		{
			if ( Config->CheckLogChatNickNameFilter(m_sNick) )
			{
				s = m_sNick;

				if ( Config->GetLogChatOption(elcoAPPENDHUBNAME) )
					s += "_" + m_pClient->GetHubName();

				if ( Config->GetLogChatOption(elcoAPPENDHUBHOST) )
					s += "_" + m_pClient->GetIP();

				if ( Config->GetLogChatOption(elcoAPPENDDATE) )
 					s += "_" + m_sTimeStamp;

				// damn hacking ;-)
				s = s.Replace('/',"");
				s = s.Replace('\\',"");
				s = s.Replace(':',"");

				s = Config->GetChatLogPath() + s + ".html";

				QFile f(s.Data());

				if ( f.open(IO_WriteOnly | IO_Append) == TRUE )
				{
					f.writeBlock(message.Data(),message.Length());
					f.close();
				}
			}
		}
	}
}

/** */
void DCChat::AddStatus( CString message, bool show )
{
	CString timestamp;

	if ( (Config->GetShowStatusMessage() == TRUE) || (show) )
	{
		timestamp = GetTimeStamp();

		// convert newline
		message = message.Replace( "\\r\xa", "\xa" );

#if QT_VERSION < 304
		if ( message.Find('\xa',0) == -1 )
			message += "<br />";
#endif
		message = message.Replace( "\xa", "<br />" );

		if ( m_bPrivateChat )
			message = "<font color=\"" + Config->GetChatColor(eccCHATSTATUSTEXT) + "\">" + message + "</font>";
		else
			message = "<font color=\"" + Config->GetChatColor(eccPUBLICCHATSTATUSTEXT) + "\">" + message + "</font>";

		if ( m_bPrivateChat )
			message = "<font color=\"" + Config->GetChatColor(eccCHATSTATUSNICK) + "\">" + "<b>&lt;DCGUI&gt;</b> " + "</font>" + message;
		else
			message = "<font color=\"" + Config->GetChatColor(eccPUBLICCHATSTATUSNICK) + "\">" + "<b>&lt;DCGUI&gt;</b> " + "</font>" + message;

		AddOutput( timestamp + message );
	}
}

/** */
void DCChat::AddMessage( CMessagePrivateChat * msg, bool bremote, bool forward )
{
	if ( msg->m_eSecureState != m_eSecureState )
	{
		SetCrypt(msg->m_eSecureState);
	}

	if ( msg->m_sSrcNick == "" )
		AddMessage( msg->m_sSrcNick.Data(), msg->m_sMessage.Data(), bremote, forward );
	else
		AddMessage( msg->m_sMultiSrcNick.Data(), msg->m_sMessage.Data(), bremote, forward );
}

/** */
void DCChat::AddMessage( CString nick, CString message, bool bremote, bool forward )
{
	unsigned int len;
	bool b;
	CString msg,s,s1,timestamp;
	int index;
	int i1,i2;
	bool b1,b2;

	//printf("ADD MESSAGE '%s'\n",message.Data());

	if ( message == "" )
	{
		return;
	}

	if ( CheckForData(message) == TRUE )
	{
		return;
	}

	// convert newline
	message = message.Replace( "\\r\xa", "\xa" );
	message = message.Replace( "\\r", "\xa" );

	// convert special chars ... damn hacking ;-)
	message = message.Replace( "<", "&lt;" );
	message = message.Replace( ">", "&gt;" );

	// convert special chars in nick ... damn hacking ;-)
	nick = nick.Replace( "<", "&lt;" );
	nick = nick.Replace( ">", "&gt;" );

	msg = message;
	s1 = "";

	b1 = b2 = TRUE;

	while ( msg != "" )
	{
		if (b1)
		{
			if ( (i1 = FindFirstLink(msg)) == -1 )
			{
				b1 = FALSE;
			}
		}
		else
		{
			i1 = -1;
		}

		if (b2)
		{
			if ( (i2 = FindFirstEmoticon(msg)) == -1 )
			{
				b2 = FALSE;
			}
		}
		else
		{
			i2 = -1;
		}

		if ( ((i1 < i2) || (i2 == -1)) && (i1 != -1) )
		{
			// convert link
			s1 += msg.Left(i1);
			msg = msg.Right(msg.Length()-i1);

			index = ConvertLinks( msg, s );
		}
		else if ( ((i2 < i1) || (i1 == -1)) && (i2 != -1) )
		{
			// convert emot
			s1 += msg.Left(i2);
			msg = msg.Right(msg.Length()-i2);

			index = ConvertEmoticons( msg, s );
		}
		else
		{
			// no link & no emot found
			s1 += msg;
			msg = "";
			break;
		}

		if ( index > 1 )
		{
			s1 += s;
			msg = msg.Right(msg.Length()-index);
			continue;
		}

		if ( index == 1 )
		{
			s1 += msg.Left(1);
		}

		msg = msg.Right(msg.Length()-index);
	}

	message = s1;

	// convert spaces
	s = "";
	b = TRUE;

	for(len=0;len<message.Length();len++)
	{
		if ( message.Data()[len] == '\xa' )
			b = TRUE;
		else if ( message.Data()[len] != ' ' )
			b = FALSE;

		if ( message.Data()[len] == ' ' )
		{
			if (b)
				s += "&nbsp;";
			else
				s += " ";
			b = !b;
		}
		else
		{
			s += message.Data()[len];
		}
	}

	message = s;

	// convert newline
#if QT_VERSION < 304
	if ( message.Find('\xa',0) == -1 )
		message += "<br />";
#endif

	message = message.Replace( "\xa", "<br />" );

	// set color
	if ( m_bPrivateChat )
		if ( bremote && (nick != m_pClient->GetNick().Data()) )
			message = "<font color=\"" + Config->GetChatColor(eccCHATREMOTETEXT) + "\">" + message + "</font>";
		else
			message = "<font color=\"" + Config->GetChatColor(eccCHATLOCALTEXT) + "\">" + message + "</font>";
	else
		if ( forward )
			message = "<font color=\"" + Config->GetChatColor(eccPUBLICPRIVATECHATTEXT) + "\">" + message + "</font>";
		else if ( bremote && (nick != m_pClient->GetNick().Data()) )
			message = "<font color=\"" + Config->GetChatColor(eccPUBLICCHATREMOTETEXT) + "\">" + message + "</font>";
		else
			message = "<font color=\"" + Config->GetChatColor(eccPUBLICCHATLOCALTEXT) + "\">" + message + "</font>";

	if ( m_bPrivateChat )
	{
		if ( nick != m_pClient->GetNick().Data() )
			message = "<font color=\"" + Config->GetChatColor(eccCHATREMOTENICK) + "\">" + "<b>" + nick + "</b>: " + message + "</font>";
		else
			message = "<font color=\"" + Config->GetChatColor(eccCHATLOCALNICK) + "\">" + "<b>" + nick + "</b>: " + message + "</font>";
	}
	else
	{
		if ( forward )
			message = "<font color=\"" + Config->GetChatColor(eccPUBLICPRIVATECHATNICK) + "\">" + "<b>&lt;" + nick + "&gt;</b> " + "</font>" + message;
		else if ( nick != m_pClient->GetNick().Data() )
			message = "<font color=\"" + Config->GetChatColor(eccPUBLICCHATREMOTENICK) + "\">" + "<b>&lt;" + nick + "&gt;</b> " + "</font>" + message;
		else
			message = "<font color=\"" + Config->GetChatColor(eccPUBLICCHATLOCALNICK) + "\">" + "<b>&lt;" + nick + "&gt;</b> " + "</font>" + message;
	}

	timestamp = GetTimeStamp();

	AddOutput( timestamp + message );

	if ( bremote && m_bPrivateChat )
	{
		Config->PlaySound(eusRECEIVE);

		// send away message
		if ( Config->GetAwayMode() == euamAWAY )
		{
			s = Config->GetAwayMessage().Data();

			if ( (s != "") && (bSendAway) )
			{
				m_pClient->SendPrivateMessage( m_pClient->GetNick(), m_sNick, s );
				bSendAway = FALSE;
			}
		}
		else
		{
			bSendAway = TRUE;
		}
	}
}

/** */
void DCChat::SendMessage()
{
	int err;
	CString message;
	CString *ps;
	QTextCodec *codec;
	CString s;

	// use text encoder
	s = Config->GetChatTextCodec( ectcINPUT );

	if ( (s != "") && ( ( codec = QTextCodec::codecForName(s.Data())) != 0) )
	{
		message = (const char*)codec->fromUnicode(TextEdit_CHATINPUT->text());
	}
	else
	{
		message = TextEdit_CHATINPUT->text().ascii();
	}

	if ( message != "" )
	{
		// add history
		if ( m_History.Count() == 10 )
		{
			ps = m_History.Next(0);
			m_History.Del(ps);
		}

		m_pCurrentHistory = 0;
		m_History.Add(new CString(message));

		if ( (CheckForCommand() == FALSE) &&
		     (m_pClient) )
		{
			// convert the newlines
			message = message.Replace( "\\r\xa", "\xa" );
			message = message.Replace( "\xa", "\xd\xa" );

			if ( m_bPrivateChat )
				err = m_pClient->SendPrivateMessage( m_pClient->GetNick(), m_sNick, message );
			else
				err = m_pClient->SendChat( m_pClient->GetNick(), message );

			if ( err == 0 )
			{
				if ( m_bPrivateChat )
					AddMessage( m_pClient->GetNick(), message, FALSE );
				Config->PlaySound(eusSEND);

				TextEdit_CHATINPUT->setText("");
			}
			else
			{
				AddStatus( tr("Message not send !").ascii() );
			}
		}
		else
		{
			TextEdit_CHATINPUT->setText("");
		}
	}
}

/** */
int DCChat::FindFirstLink( CString msg )
{
	int i,i1,i2;

	for(i=0,i1=-1,i2=-1;p_LinkList[i]!=0;i++)
	{
		if ( (i1 = msg.Find(p_LinkList[i],0,FALSE)) != -1 )
		{
			if ( (i2 == -1) || (i2 > i1) )
			{
				i2 = i1;
			}
		}
	}

	return i2;
}

/** */
int DCChat::ConvertLinks( CString msg, CString & s )
{
	int ret = 1;
	int i,i1,i2,i3;

	// convert links
	for(i=0;p_LinkList[i]!=0;i++)
	{
		if ( msg.Left(strlen(p_LinkList[i])).ToUpper() == p_LinkList[i] )
		{
			// link end at space or newline
			i1 = msg.Find(' ',0,FALSE);
			i2 = msg.Find('\xa',0,FALSE);
			i3 = msg.Find('\xd',0,FALSE);

			if ( i1 == -1 )
				i1 = msg.Length();
			if ( (i1 > i2) && (i2 != -1) )
				i1 = i2;
			if ( (i1 > i3) && (i3 != -1) )
				i1 = i3;

			s = msg.Left(i1);
			s = "<a title=\""+s+"\" name=\""+s+"\" href=\""+s+"\">"+s+"</a>";

			ret = i1;

			break;
		}
	}

	return ret;
}

/** */
int DCChat::FindFirstEmoticon( CString msg )
{
	int i1,i2;
	CList<DC_EmoticonObject> * elist;

	elist = Config->EmoticonList();

	i1 = i2 = -1;

	if ( (Config->GetEnableEmoticons() == TRUE) && (elist != 0) )
	{
		DC_EmoticonObject * EmoticonObject = 0;

		while ( (EmoticonObject=elist->Next(EmoticonObject)) != 0 )
		{
			if ( (i1 = msg.Find( EmoticonObject->m_Text, 0 )) != -1 )
			{
				if ( (i2 == -1) || (i2 > i1) )
				{
					i2 = i1;
				}
			}
		}
	}

	return i2;
}

/** */
int DCChat::ConvertEmoticons( CString msg, CString & s )
{
	int ret = 1;
	CList<DC_EmoticonObject> * elist;

	elist = Config->EmoticonList();

	// convert emoticons
	if ( (Config->GetEnableEmoticons() == TRUE) && (elist != 0) )
	{
		DC_EmoticonObject * EmoticonObject = 0;

		while ( (EmoticonObject=elist->Next(EmoticonObject)) != 0 )
		{
			if ( msg.Left(EmoticonObject->m_Text.Length()) == EmoticonObject->m_Text )
			{
				s   = "<img alt=\"\" align=\"center\" source=\"emoticon"+CString().setNum(EmoticonObject->m_nID)+"\" />";
				ret = EmoticonObject->m_Text.Length();
				break;
			}
		}
	}

	return ret;
}

/** */
bool DCChat::CheckForCommand()
{
	bool res = FALSE;
	QString timestamp;
	CString cmd,s;

	cmd = TextEdit_CHATINPUT->text().ascii();

	if ( cmd != "" )
	{
		// help
		if ( cmd == "/dchelp" )
		{
			s  = tr("Help:").ascii();
			s += "\n";
			s += tr("/clear - clear the chat window").ascii();
			s += "\n";
			s += tr("/mode - show current mode").ascii();
			s += "\n";
			s += tr("/refresh - refresh share").ascii();
			s += "\n";
			s += tr("/ts - switch time display in chat on/off").ascii();
			s += "\n";
			s += tr("/dcgui - send an advertisment to the hub").ascii();
			s += "\n";
			s += tr("/join &lt;address&gt; - disconnect from currently connected hub and connect to another hub").ascii();
			s += "\n";
			s += tr("/msg &lt;nick&gt; &lt;message&gt; - send private message").ascii();
			s += "\n";
			s += tr("/away &lt;message&gt; - set automatic response for private messages; if the message is empty, disable it").ascii();

			AddStatus( s, TRUE );
			res = TRUE;
		}
		// /clear - clear the chat window
		else if ( cmd == "/clear" )
		{
			TextEdit_CHATOUTPUT->clear();

			AddStatus( tr("Chat Cleared.").ascii(), TRUE );
			res = TRUE;
		}
		// /mode - show current mode
		else if ( cmd == "/mode" )
		{
			s  = tr("Current mode: ");

			if ( Config->GetMode() == ecmACTIVE )
				s += tr("active");
			else
				s += tr("passive");
			s += ".";

			AddStatus( s, TRUE );
			res = TRUE;
		}
		// /refresh - refresh share 
		else if ( cmd == "/refresh" )
		{
			if ( CFileManager::Instance()->CreateShareList() == TRUE )
			{
				s = tr("Refresh share in progress.").ascii();
			}
			else
			{
				s = tr("Refresh share allready in progress.").ascii();
			}

			AddStatus( s, TRUE );
			res = TRUE;
		}
		// /ts - switch time display in chat on/off 
		else if ( cmd == "/ts" )
		{
			bool b;

			if ( m_bPrivateChat )
			{
				b = !Config->GetTimeStamp(etsPRIVATECHAT);
				Config->SetTimeStamp(etsPRIVATECHAT,b);
			}
			else
			{
				b = !Config->GetTimeStamp(etsHUBCHAT);
				Config->SetTimeStamp(etsHUBCHAT,b);
			}

			s = tr("Switch timestamp ").ascii();

			if ( b )
				s += "on.";
			else
				s += "off.";

			AddStatus( s, TRUE );
			res = TRUE;
		}
		// /dcgui - send an advertisment to the hub
		else if ( cmd == "/dcgui" )
		{
			AddStatus( tr("Send advertisment.").ascii(), TRUE );

			s = tr("\nDownload DCGUI-QT from http://dc.ketelhot.de\nIt does everything, and makes you coffee\n");

			m_pClient->SendChat( m_pClient->GetNick(), s );

			res = TRUE;
		}
		// /join <address> - disconnect from currently connected hub and connect to another hub 
		else if ( cmd.Find("/join ",0,FALSE) == 0 )
		{
			CString address;

			address = cmd.Mid(6,cmd.Length()-6);

			if ( address == "" )
				s = tr("Wrong parameter for '/join'").ascii();
			else
			{
				s = tr("Join to: ").ascii() + address;

				m_pClient->SetHubName(address);
				m_pClient->setCaption(address.Data());
				m_pClient->Connect(address);
			}

			AddStatus( s, TRUE );
			res = TRUE;
		}
		// /msg <nick> <message> - send private message 
		else if ( cmd.Find("/msg ",0,FALSE) == 0 )
		{
			int i;
			CString nick,message="";

			cmd = cmd.Mid(5,cmd.Length()-5);

			if ( (i = cmd.Find(' ')) != -1 )
			{
				nick    = cmd.Left(i);
				message = cmd.Mid(i+1,cmd.Length()-i-1);
			}

			if ( (nick == "") || (message == "") )
			{
				s = tr("Wrong parameter for '/msg'").ascii();
			}
			else
			{
				if ( m_pClient->SendPrivateMessage( m_pClient->GetNick(), nick, message ) == 0 )
				{
					s = tr("Private message send.").ascii();
					Config->PlaySound(eusSEND);
				}
				else
				{
					s = tr("Private message not send !").ascii();
				}
			}

			AddStatus( s, TRUE );
			res = TRUE;

		}
		// /away <message> - set automatic response for private messages; if the message is empty, disable it 
		else if ( cmd.Find("/away",0,FALSE) == 0 )
		{
			int i;

			s = "";

			i = cmd.Find(' ',5);

			if ( i > 0 )
				s = cmd.Mid(i+1,cmd.Length()-i-1);

			Config->SetAwayMessage(s);

			if ( s == "" )
				s = tr("Away message disabled.").ascii();
			else
				s = tr("Away message set.").ascii();

			AddStatus( s, TRUE );
			res = TRUE;
		}
		else if ( cmd.Find("/sendphoto",0,FALSE) == 0 )
		{
			if ( !m_bPrivateChat )
			{
				s = tr("Only allowed in private chat.").ascii();
			}
			else if ( m_pClient->GetUserClientVersion(m_sNick) != eucvDCGUI )
			{
				s = tr("Client use wrong version.").ascii();
			}
			else
			{
				CString fn = Config->GetUserPhotoFileName();
				CDir dir;
				CFile f;
				CByteArray bas,bad;
				ulonglong size;
				CBase64 base64;

				size = dir.getFileSize( fn, FALSE );

				if ( (size == 0) || (f.Open(fn,IO_READONLY|IO_RAW) == FALSE) )
				{
					s = tr("No Photo found.").ascii();
				}
				else
				{
					bas.SetSize(size);

					if ( f.Read((char*)bas.Data(),size) != size )
					{
						s = tr("Can't read photo.").ascii();
					}
					else
					{
						base64.Encode( &bad, &bas );

						if ( bad.Size() == 0 )
						{
							s = tr("Photo encode error.").ascii();
						}
						else
						{
							CString se;

							se.Set((char*)bad.Data(),bad.Size());
							se  = "<photo data=\"" + se + "\">";

							if ( m_pClient->SendPrivateMessage( m_pClient->GetNick(), m_sNick, se ) != 0 )
							{
								s = tr("Photo not send.").ascii();
							}
							else
							{
								s = tr("Photo send.").ascii();
							}
						}
					}
				}

				f.Close();
			}

			AddStatus( s, TRUE );
			res = TRUE;
		}
	}
	return res;
}

/** */
bool DCChat::CheckForData( CString message )
{
	bool res = FALSE;

	// now we check for special private messages
	if ( m_bPrivateChat )
	{
		// remote send a photo
		if ( (message.Find("<photo ") == 0) && (message.Find(">",7) > 0) )
		{
			int i1,i2;
			CString s;
			CByteArray ba;
			CBase64 base64;

			if ( (i1 = message.Find("data=\"")) != -1 )
			{
				if ( (i2 = message.Find("\"",i1+6)) != -1 )
				{
					//printf("'%s'\n",message.Data());
					//printf("'%s'\n",message.Mid(i1+6,i2-i1-6).Data());

					s = message.Mid(i1+6,i2-i1-6);

					if ( base64.Decode( &ba, &s ) > 0 )
					{
						pUsersList->AddFriendPhoto(m_sNick,&ba);
						AddStatus(tr("Photo received.").ascii(), TRUE );

						res = TRUE;
					}
				}
			}
		}
	}

	return res;
}
